import string
from FiniteStateMachine import FiniteStateMachine

Keywords = ['as', 'class', 'break',	'continue',	'do', 'else',
    'for', 'fun', 'false', 'if', 'in', 'interface',
    'super', 'return', 'object', 'package',	'null',	'is',
    'try', 'throw', 'true', 'this', 'typeof', 'typealias',
    'when',	'while', 'val',	'var']

Operators = ['+', '-', '*', '/', '%', '=',
    '+=', '-=', '*=', '/=', '%=',
    '++', '--',
    '&&', '||', '!',
    '==', '!=',
    '===', '!==',
    '<', '>', '<=', '>='
    '[', ']', '!!', '?.', '?:', '::', '..', ':', '?', '->', '@', ';', '$', '_']

Punctuators = [';', ':', ',', '.', '(', ')', '[', ']', '{', '}']

def isMultilineCommentStart(str):
    return str.startswith('/*')

def isMultilineCommentEnd(str):
    return str.startswith('*/')

def isMultilineComment(str):
    if isMultilineCommentStart(str):
        pos=2
        while pos < len(str) and not isMultilineCommentEnd(str[pos:]):
            pos+=1
        if pos < len(str)-1 and isMultilineCommentEnd(str[pos:]):
            return pos+2, "CommentMultiline", str[:pos+2]
        return pos+2, "UnfinishedComment", str
    return 0, None, None

def isCommentStart(str):
    return str.startswith('//')

def isLineEnd(str):
    return str[0] == '\n'

def isComment(str):
    if isCommentStart(str):
        pos=2
        while pos < len(str):
            if isLineEnd(str[pos:]):
                return pos, "Comment", str[:pos]
            pos+=1
        return len(str), "Comment", str
    return 0, None, None

def isWhiteSpace(str):
    if str[0] == '\n':
        return 1, "Newline", str[:1]
    if str[0] == ' ':
        return 1, "Whitespace", str[:1]
    if str[0]=='\t':
        return 1, "Whitespace", str[:1]
    if str[0] in string.whitespace:
        return 1, "Whitespace", str[:1]
    return 0, None, None

def isPunctuator(str):
    if str[0] in Punctuators:
        return 1, "Punctuator", str[:1]
    else:
        return 0, None, None

def isIdentifier(str):
    if str[0] == '`':
        pos=0
        while pos < len(str) and str[pos]!='`':
            pos+=1
        return pos, 'Identifier', str[:pos]
    if str[0].isalpha() or str[0]=='_':
        pos = 0
        while pos < len(str) and (str[pos].isalpha() or str[pos].isdigit() or str[pos]=='_'):
            pos+=1
        return pos, "Identifier", str[:pos]
    return 0, None, None

def isKeyword(str):
    for k in Keywords:
        if str.startswith(k):
            if len(k) >= len(str) or (not str[len(k)].isalnum() and not str[len(k)] in ['_']):
                return len(k), "Keyword", str[:len(k)]
    return 0, None, None

def isOperator(str):
    opers = [op for op in Operators if str.startswith(op)]
    opers.sort(key=lambda a: -len(a))
    if len(opers):
        return len(opers[0]), "Operator", str[:len(opers[0])]
    else:
        return 0, None, None

def isNullLiteral(str):
    if len(str) >= 3 and str.startswith('null'):
        if len(str) == 3 or (not str[4].isalnum() and not str[4] in ['_']):
            return 4, "Literal", str[:4]
    return 0, None, None

def isBooleanLiteral(str):
    if len(str) >= 4 and str.startswith('true'):
        if len(str) == 4 or (not str[5].isalnum() and not str[5] in ['_']):
            return 5, "Literal", str[:5]
    if len(str) >= 5 and str.startswith('false'):
        if len(str) == 5 or (not str[6].isalnum() and not str[6] in ['_']):
            return 6, "Literal", str[:6]
    return 0, None, None

def isCharacterLiteral(str):
    if (len(str) >= 7 and str[0]=='\''
       and ( str[1]=='u' or str[1]=='x' )
       and (str[2].isdigit() and str[3].isidigit() and str[4].isdigit() and str[5].isdigit())
       and str[6] == '\''):
        return 7, "Literal", str[:7]
    if len(str) >= 4 and str[0]=='\'' and str[1]=='\\' and str[3]=='\'':
        return 4, "Literal", str[:4]
    if len(str) >= 3 and str[0]=='\'' and str[2]=='\'':
        return 3, "Literal", str[:3]
    return 0, None, None

def isStringLiteral(str):
    # todo: String Interpolation
    if len(str) >= 2:
        pos = 0
        Interpolation = False
        if str[pos]=='$':
            pos+=1
            Interpolation=True
        if str[pos] == '"':
            if len(str) >= 6:
                if str[pos+1] == '"' and str[pos+2] == '"':
                    # multiline string
                    pos+=3
                    while pos < len(str) - 2:
                        if str[pos] == '"' and str[pos+1] == '"' and str[pos+2] == '"':
                            return pos+3, "Literal", str[:pos+3]
                        pos+=1
                    return pos+2, "Literal", str
            while pos < len(str):
                if pos + 1 < len(str):
                    if (not str[pos]=='\\' and str[pos+1]=='"'):
                        return pos+2, "Literal", str[:pos+2]
                pos += 1
    return 0, None, None

class DecLiteralFSM(FiniteStateMachine):
    def __init__(self):
        super().__init__()
        self.pos=0
        self.add_state('s0', self.s0_handler)
        self.add_state('s1', self.s1_handler)
        self.add_state('s2', self.s2_handler)
        self.add_state('s3', self.s3_handler)
        self.add_state('s4', self.s4_handler)
        self.add_state('s5', self.s5_handler)
        self.add_state('s6', self.s6_handler)
        self.add_state('passed', self.passed_handler, True)
        self.add_state('exit_state', self.exit_handler, True)
        self.set_start('s0')

    def run(self, input):
        return super().run(input), self.pos

    def s0_handler(self, str):
        if str[self.pos].isdigit():
            self.pos+=1
            return 's1', str
        return 'exit_state', str

    def s1_handler(self, str):
        if str[self.pos].isdigit() or str[self.pos] == '_':
            self.pos+=1
            return 's1', str
        if str[self.pos] == '.' and not (self.pos+1 < len(str) and str[self.pos+1] == '.'):
            self.pos+=1
            return 's2', str
        if str[self.pos] == 'E' or str[self.pos] == 'e':
            self.pos+=1
            return 's4', str
        return 'passed', str

    def s2_handler(self, str):
        if str[self.pos].isdigit() or str[self.pos] == '_':
            self.pos+=1
            return 's3', str
        return 's3', str

    def s3_handler(self, str):
        if str[self.pos].isdigit() or str[self.pos] == '_':
            self.pos+=1
            return 's3', str
        if str[self.pos] == 'E' or str[self.pos] == 'e':
            self.pos+=1
            return 's4', str
        if str[self.pos].isalpha():
            self.pos+=1
            return 's3', str
        return 'passed', str

    def s4_handler(self, str):
        if str[self.pos] == '+' or str[self.pos] == '-':
            self.pos+=1
            return 's5', str
        if str[self.pos].isdigit() or str[self.pos] == '_':
            self.pos+=1
            return 's6', str
        return 'exit_state', str

    def s5_handler(self, str):
        if str[self.pos].isdigit() or str[self.pos] == '_':
            self.pos+=1
            return 's6', str
        return 'exit_state', str

    def s6_handler(self, str):
        if str[self.pos].isdigit() or str[self.pos] == '_':
            self.pos+=1
            return 's6', str
        return 'passed', str

    def passed_handler(self, str):
        return None, str
    def exit_handler(self, str):
        return None, str

class HexLiteralFSM(FiniteStateMachine):

    pos=0

    def __init__(self):
        super().__init__()
        self.add_state('s0', self.s0_handler)
        self.add_state('s1', self.s1_handler)
        self.add_state('s2', self.s2_handler)
        self.add_state('s3', self.s3_handler)
        self.add_state('passed', self.passed_handler, True)
        self.add_state('exit_state', self.exit_handler, True)
        self.set_start('s0')

    def run(self, input):
        return super().run(input), self.pos

    def s0_handler(self, str):
        if str[self.pos] == '0':
            self.pos+=1
            return 's1', str
        return 'exit_state', str

    def s1_handler(self, str):
        if str[self.pos] == 'x':
            self.pos+=1
            return 's2', str
        return 'exit_state', str

    def s2_handler(self, str):
        if str[self.pos].isdigit() or str[self.pos] in ['A', 'B', 'C', 'D', 'E', 'F'] or str[self.pos] == '_':
            self.pos+=1
            return 's3', str
        return 'exit_state', str

    def s3_handler(self, str):
        if str[self.pos].isdigit() or str[self.pos] in ['A', 'B', 'C', 'D', 'E', 'F'] or str[self.pos] == '_':
            self.pos+=1
            return 's3', str
        return 'passed', str

    def passed_handler(self, str):
        return None, str
    def exit_handler(self, str):
        return None, str

class BinLiteralFSM(FiniteStateMachine):

    pos=0

    def __init__(self):
        super().__init__()
        self.add_state('s0', self.s0_handler)
        self.add_state('s1', self.s1_handler)
        self.add_state('s2', self.s2_handler)
        self.add_state('s3', self.s3_handler)
        self.add_state('passed', self.passed_handler, True)
        self.add_state('exit_state', self.exit_handler, True)
        self.set_start('s0')

    def run(self, input):
        return super().run(input), self.pos

    def s0_handler(self, str):
        if str[self.pos] == '0':
            self.pos+=1
            return 's1', str
        return 'exit_state', str

    def s1_handler(self, str):
        if str[self.pos] == 'b':
            self.pos+=1
            return 's2', str
        return 'exit_state', str

    def s2_handler(self, str):
        if str[self.pos] in ['0', '1'] or str[self.pos] == '_':
            self.pos+=1
            return 's3', str
        return 'exit_state', str

    def s3_handler(self, str):
        if str[self.pos] in ['0', '1'] or str[self.pos] == '_':
            self.pos+=1
            return 's3', str
        return 'passed', str

    def passed_handler(self, str):
        return None, str
    def exit_handler(self, str):
        return None, str

def isNumericLiteral(str):
    fsm = DecLiteralFSM()
    state, pos = fsm.run(str)
    if state == 'passed':
        return pos, "Literal", str[:pos]
    fsm = HexLiteralFSM()
    state, pos = fsm.run(str)
    if state == 'passed':
        return pos, "Literal", str[:pos]
    fsm = BinLiteralFSM()
    state, pos = fsm.run(str)
    if state == 'passed':
        return pos, "Literal", str[:pos]
    return 0, None, None


def countNewlines(str):
    counter=0
    for s in str:
        if s=='\n':
            counter+=1
    return counter

class TokenData:
    def __init__(self, token, str, line):
        self.token = token
        self.str = str
        self.line = line

    def __str__(self):
        return self.token + ": " + self.str

    def __repr__(self):
        return self.token + ": " + self.str

class KotlinLexer:

    pos=0
    line=0

    recognizers = [ isWhiteSpace,
                    isComment,
                    isMultilineComment,
                    isKeyword,
                    isIdentifier,
                    isPunctuator,
                    isOperator,
                    isNullLiteral,
                    isBooleanLiteral,
                    isStringLiteral,
                    isCharacterLiteral,
                    isNumericLiteral ]

    def __init__(self, str=None):
        self.str = str
        self.reset()


    def nextToken(self):
        if self.pos >= len(self.str):
            return TokenData("EndOfInput", "", 0)
        if set(self.str[self.pos]).difference(string.printable):
            self.pos+=1
        if self.pos >= len(self.str):
            return TokenData("EndOfInput", "", 0)
        for recognizer in self.recognizers:
            delta_pos, token_type, value = recognizer(self.str[self.pos:])
            if delta_pos != 0:
                self.line+=countNewlines(self.str[self.pos:self.pos+delta_pos])
                self.pos += delta_pos
                return TokenData(token_type, value, self.line)
        return TokenData("Error", "", 0)

    def get_tokens(self, str=None):
        if str!=None:
            self.str = str
        self.reset()
        tokens=[]
        token_data = self.nextToken()
        while token_data.token != "EndOfInput" and token_data.token != "Error":
            tokens.append(token_data)
            token_data = self.nextToken()
        return tokens

    def reset(self):
        self.pos = 0
        self.line = 0



class FSMState:
    def __init__(self, tokenType):
        self.tokenType = tokenType

    def isEndState(self):
        return not self.tokenType == None

class FSMHandler:

    def __init__(self, states, transitions, defaultState):
        # states - array, transitions - array of array of strings
        self.transitions=transitions
        self.states=states
        self.defaultState=defaultState

    def __call__(self, str):
        for i, transition in enumerate(transitions):
            result = transition(str)
            if result:
                return states[i], result
        return defaultState, str[1:]

class FiniteStateMachine:
    def __init__(self):
        self.handlers = {}
        self.startState = None
        self.endStates = []

    def add_state(self, name, handler, end_state=0):
        name = name.upper()
        self.handlers[name] = handler
        if end_state:
            self.endStates.append(name)

    def set_start(self, name):
        self.startState = name.upper()

    def run(self, input):
        try:
            handler = self.handlers[self.startState]
        except:
            raise InitializationError("must call .set_start() before .run()")
        if not self.endStates:
            raise  InitializationError("at least one state must be an end_state")

        while True:
            (newState, input) = handler(input)
            if newState.upper() in self.endStates:
                return newState
            else:
                handler = self.handlers[newState.upper()]



from distutils.core import setup
import re
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = re.search(
    '^__version__\s*=\s*"(.*)"',
    open('kotlin_style_check/KotlinStyleChecker.py').read(),
    re.M
).group(1)

setup(
  name = 'kotlin_style_check',       
  packages=setuptools.find_packages(),
  version = version,     
  license='MIT',  
  description = 'Library that provides ability to fix naming and comments in Kotlin code files according to Kotlin code conventions',
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = 'Vermillio',     
  author_email = 'morganas.scream@gmail.com',     
  url = 'https://github.com/Vermillio/Metaprogramming2.0',
  keywords = ['code conventions', 'formatting', 'naming', 'Kotlin', 'C#', 'lexical analysis', 'parsing'],
  classifiers=[
    'Development Status :: 3 - Alpha',    
    'Intended Audience :: Developers',   
    'License :: OSI Approved :: MIT License',
  ],
  python_requires='>=3.6',
)


from kotlin_style_check.KotlinLexer import KotlinLexer, TokenData
import os
import argparse
import fnmatch
import logging
import re

__version__ = "0.6"

class KotlinStyleChecker:
    def __init__(self):
        self.fixed_name_str = 'line {}: {} -> {} - Fixed {} name'
        self.logger = logging.getLogger('KotlinStyleCheckerLogger')

    ''' if this is not called then no logging happens '''
    def setup_logger(self, log_file, level=logging.INFO):
        logger = logging.getLogger('KotlinStyleCheckerLogger')
        logger.setLevel(level)
        formatter = logging.Formatter('%(message)s')
        handler = logging.FileHandler(log_file)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    def _log_name_fixed(self, line, old_str, new_str, log_name):
        if old_str != new_str:
            self.logger.info(self.fixed_name_str.format(line, old_str, new_str, 'package'))
        return old_str, new_str

    def to_camel_case(self, str, first_capital=False, leave_underscores=False):
        str = reversed([
                str[i].upper() if (i>0 and ((str[i].isupper() and str[i-1].islower())) or not str[i-1].isalpha())
                else str[i].lower() if str[i].isalpha()
                else str[i] for i in reversed(range(len(str))) ])
        str = [ s for s in str if s.isalnum() or (leave_underscores == True and s=='_')]
        if str[0].isalpha():
                str[0] = str[0].upper() if first_capital else str[0].lower()
        return ''.join(str)

    def to_scream_snake_case(self, str):
        for i in reversed(range(1,len(str))):
            if str[i].isupper() and str[i-1].islower():
                str = str[:i]+'_'+str[i:]
        return ''.join([s.upper() if s.isalpha() else s for s in str])

    def _remove_whitespaces(self, tokens):
        return [t for t in tokens if t.token not in ["Whitespace", 'Newline']]

    def fix_multiline_comment(self, str, line):
        if not '\n' in str:
            return str, str
        new_str = str if str[2] == '*' else str[:2] + '*' + str[2:]
        new_str = new_str if new_str[3] == '\n' else new_str[:3] + '\n' + new_str[3:]
        new_str = re.sub('\n[ \t]*\*?[ \t]?', '\n * ', new_str)
        new_str = new_str if new_str[-3] == ' ' else new_str[:-2]+' '+new_str[-2:]
        new_str = new_str if new_str[-4] == '\n' else new_str[:-3]+'\n'+new_str[-3:]
        if '@param' in new_str or '@return' in new_str:
            self.logger.warning("line {}: Avoid using @param and @return tags in comments.".format(line))
        if str != new_str:
            self.logger.info('line {}: Fixed comment marking\n\n{}\n->\n{}\n\n'.format(line, str, new_str))
        return str, new_str

    def fix_class_name(self, token):
        return self._log_name_fixed(token.line, token.str, self.to_camel_case(token.str, first_capital=True), 'class')

    def fix_package_name(self, token):
        return self._log_name_fixed(token.line, token.str, self.to_camel_case(token.str), 'package')

    def fix_function_name(self, token):
        if not (token.str[0] == '`' and token.str[-1] == '`'):
            str = self.to_camel_case(token.str)
        elif token.str[1:-1] not in Keywords:
                self.logger.info('line {}: {} - Backticks are used in method name and method name is not a reserved keyword. Refactoring is recommended.'.format(token.line, token.str))
        return self._log_name_fixed(token.line, token.str, str, 'function')

    def fix_test_method_name(self, token):
        if not (token.str[0] == '`' and token.str[-1] == '`'):
            return self._log_name_fixed(token.line, token.str, self.to_camel_case(token.str, leave_underscores=True), 'test method')
        return token.str, token.str

    def fix_const_var_name(self, token):
        str = self.to_scream_snake_case(token.str)
        return self._log_name_fixed(token.line, token.str, str, 'const variable')

    def fix_var_name(self, token):
        return self._log_name_fixed(token.line, token.str, self.to_camel_case(token.str), 'variable')

    def fix_private_var_name(self, token):
        return self._log_name_fixed(token.line, token.str, '_'+self.to_camel_case(token.str), 'private variable')

    def check_comments(self, tokens, replacements):
        for token in tokens:
            if token.token == 'CommentMultiline':
                str_from, str_to = self.fix_multiline_comment(token.str, token.line)
                if str_from != str_to:
                    replacements[str_from] = str_to

    def check_naming(self, tokens, replacements):
        str_from, str_to, level = "", "", 0
        level = 0
        private_var_indices = []
        for i in range(len(tokens)):
            level += 1 if tokens[i].str == '{' else -1 if tokens[i].str == '}' else 0
            if level < 0:
                self.logger.warning('line {}: } is used without opening {')
            if i > 0 and tokens[i].token == 'Identifier' and tokens[i-1].token == 'Keyword':
                if tokens[i-1].str == 'package':
                    str_from, str_to = self.fix_package_name(tokens[i])
                elif tokens[i-1].str == 'class':
                    str_from, str_to = self.fix_class_name(tokens[i])
                elif tokens[i-1].str == 'fun' and i > 2 and tokens[i-2].str == 'Test' and tokens[i-3].str == '@':
                    str_from, str_to = self.fix_test_method_name(tokens[i])
                elif tokens[i-1].str == 'fun':
                        str_from, str_to = self.fix_function_name(tokens[i])
                elif tokens[i-1].str == 'val' and ((i<len(tokens)-1 and tokens[i+1] == '=' and level==0) or (i>2 and tokens[i-2].str == 'const')):
                    str_from, str_to = self.fix_const_var_name(tokens[i])
                elif tokens[i-1].str == 'val' and (i>2 and tokens[i-2].str == 'private'):
                    private_var_indices.append(i)
                elif tokens[i-1].str == 'val':
                    str_from, str_to = self.fix_var_name(tokens[i])
            if str_from != str_to and str_from not in replacements.keys():
                    while str_to in list(replacements.values()):
                        str_to += "_"
                    replacements[str_from] = str_to
        for i in private_var_indices:
            str_from, str_to = self.fix_private_var_name(tokens[i]) if self.to_camel_case(tokens[i].str) in list(replacements.values()) else self.fix_var_name(tokens[i])
            if str_from != str_to and str_from not in replacements.keys():
                while str_to in list(replacements.values()):
                    str_to += "_"
                replacements[str_from] = str_to

    def _replace(self, strings, replacements):
        return [ replacements[s] if s in replacements.keys() else s for s in strings ]

    def fix(self, file_contents, log_files):
        if not file_contents or not log_files or len(file_contents) == 0 or len(file_contents) != len(log_files):
            return None
        file_tokens, replacements = [], {}
        for file_content, log_file in zip(file_contents, log_files):
            self.setup_logger(log_file)
            tokens = KotlinLexer(file_content).get_tokens(file_content)
            cleared_tokens = self._remove_whitespaces(tokens)
            self.check_comments(cleared_tokens, replacements)
            self.check_naming(cleared_tokens, replacements)
            file_tokens += [tokens]
        return [''.join(self._replace([token.str for token in tokens], replacements)) for tokens in file_tokens]

    def fix_in_files(self, input_path):
        file_contents = []
        file_names = []
        if os.path.isdir(input_path):
            for dir, _, files in os.walk(input_path):
                rel_dir = os.path.relpath(dir, input_path)
                kotlin_files = fnmatch.filter(files, "*.kt")
                for file in kotlin_files:
                    file_name = os.path.join(dir, file)
                    with open(file_name) as f:
                        file_contents.append(f.read())
                        file_names.append(file_name)
        elif os.path.isfile(input_path):
            with open(input_path) as f:
                file_contents.append(f.read())
                file_names.append(input_path)
        else:
            return
        file_contents = self.fix(file_contents, log_files=[f[:-3]+"_fixing.log" for f in file_names])
        for file_name,  file_content in zip(file_names, file_contents):
            with open(file_name, 'w') as f:
                f.write(file_content)

    def run_tests(self, log_file):
        self.setup_logger(log_file)
        file_contents = ["""
        class C {
            private val elementList = mutableListOf<Element>()
            val ElementList: List<Element>
                 get() = elementList
        }""",
        """/* l\no\ng */""",
        """/* log */"""]
        expected = ["""
        class C {
            private val _elementList = mutableListOf<Element>()
            val elementList: List<Element>
                 get() = _elementList
        }""",
        """/**\n * l\n * o\n * g\n */""",
        """/* log */"""]
        got = KotlinStyleChecker().fix(file_contents, log_files=["tests.log"] * len(file_contents))
        for e, g in zip(expected, got):
            if e != g:
                self.logger.info("TEST FAILED\nExpected:\n\n{}\n\nGot:\n\n{}\n\n".format(e,g))

def main():
    parser = argparse.ArgumentParser(description="Specify input path (file or dir)")
    parser.add_argument("--input_path", help="path to check Kotlin code conventions", required=True)
    args = parser.parse_args()
    kotlin_style_checker = KotlinStyleChecker()
    kotlin_style_checker.run_tests('tests.log')
    kotlin_style_checker.fix_in_files(args.input_path)

if __name__ == "__main__":
    main()
